// AffineTransformations.cpp : This file contains the 'main' function. Program execution begins and ends there.
//

#include <opencv2/core.hpp>
#include <opencv2/imgproc.hpp>
#include <opencv2/highgui.hpp>

#include <iostream>

cv::Mat createTransformation(float angle, float tx, float ty, float sx, float sy, float skew_factor, float persp1, float persp2)
{
	// Rotation
	cv::Mat R = cv::Mat::eye(3, 3, CV_32F);
	R.at<float>(0, 0) = cos(angle);
	R.at<float>(0, 1) = -sin(angle);
	R.at<float>(1, 0) = sin(angle);
	R.at<float>(1, 1) = cos(angle);

	// Translation
	cv::Mat t = cv::Mat::eye(3, 3, CV_32F);
	t.at<float>(0, 2) = tx;
	t.at<float>(1, 2) = ty;

	// Scaling
	cv::Mat Scale = cv::Mat::eye(3, 3, CV_32F);
	Scale.at<float>(0, 0) = sx;
	Scale.at<float>(1, 1) = sy;

	// Skewing
	cv::Mat Skew = cv::Mat::eye(3, 3, CV_32F);
	Skew.at<float>(0, 1) = skew_factor;

	cv::Mat T = t * R * Skew * Scale;

	// Perspective transformation
	T.at<float>(2, 0) = persp1;
	T.at<float>(2, 1) = persp2;

	return T;
}

cv::Mat applyTransformation(cv::Mat& img, cv::Mat T, bool isPerspective)
{
	cv::Mat newImg = cv::Mat::zeros(img.size(), CV_8UC3);

	int HEIGHT = img.rows;
	int WIDTH = img.cols;

	for (int i = 0; i < HEIGHT; ++i)
	{
		for (int j = 0; j < WIDTH; ++j)
		{
			cv::Mat new_p(3, 1, CV_32F);
			new_p.at<float>(0, 0) = i;
			new_p.at<float>(1, 0) = j;
			new_p.at<float>(2, 0) = 1.f;

			cv::Mat original_p = T.inv() * new_p; // From the new image, we transform back to the original image...
			
			float Z = 1.f;
			if (isPerspective) // In case of perspective transformation, homogeneous division needed!
			{
				Z = original_p.at<float>(2, 0);
			}
			int X = original_p.at<float>(0, 0) / Z;
			int Y = original_p.at<float>(1, 0) / Z;

			if (0 <= X && X < HEIGHT && 0 <= Y && Y < WIDTH)
			{
				newImg.at<cv::Vec3b>(i, j) = img.at<cv::Vec3b>(X, Y); // ...and get the color
			}
		}
	}

	return newImg;
}

int main()
{
	cv::Mat img = cv::imread("T:\\opencv\\sources\\samples\\data\\apple.jpg");

	if (img.empty())
	{
		std::cout << "Couldn't read image!" << std::endl;
		return 0;
	}

	cv::Mat T = createTransformation(0.81, 200, 150, 0.5, 0.5, 0.2, 0.0001, 0.002);
	std::cout << "T = " << T << std::endl;
	cv::Mat newImg = applyTransformation(img, T, true);

	cv::imshow("Transformations", newImg);

	int key = cv::waitKey(0);

	return 0;
}
